@@ -498,6 +498,7 @@ msole_prop_min_size (guint32 type)
 	bytes_needed = (_n);						\
 	if (_s1 > 0 && (data_end - *data) / _s1 < bytes_needed) {	\
 		g_warning ("Invalid MS property or file truncated");	\
+		g_free (res);						\
 		return NULL;						\
 	}								\
 	bytes_needed *= _s1;						\
@@ -629,6 +630,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 			g_warning ("Unhandled property value type %d (0x%x)",
 				   type, type);
 		NEED_BYTES (8);
+		ADVANCE;
 		break;
 
 	case VT_BSTR :
@@ -2522,6 +2524,11 @@ gsf_msole_inflate (GsfInput *input, gsf_off_t offset)
 /*				fprintf (stderr, "Shift %d, token len %d, distance %d bytes %.2x %.2x\n",
 				shift, len, distance, (token & 0xff), (token >> 8)); */
 
+				if (distance >= pos) {
+					g_warning ("Corrupted compressed stream");
+					break;
+				}
+
 				for (i = 0; i < len; i++) {
 					unsigned srcpos = (pos - distance - 1) % VBA_COMPRESSION_WINDOW;
 					guint8 c = buffer [srcpos];
